@@ -877,6 +877,20 @@ static int flyby_buttn(int b, int d)
 
 /*---------------------------------------------------------------------------*/
 
+#define TOUCH_SENSITIVITY       20
+#define LONG_PRESS_TIME 1000 /* one second */
+#define DOUBLE_TAP_TIME 500
+static enum {
+    BTN_CLEAR,
+    BTN_PRESS1,
+    BTN_CLICKED,
+    BTN_PRESS2,
+    BTN_DRAGGING
+} button_state = BTN_CLEAR;
+static int button_pressed_x = 0;
+static int button_pressed_y = 0;
+static Uint32 button_time = 0;
+
 static int stroke_rotate = 0;
 static int stroke_mag    = 0;
 
@@ -912,6 +926,18 @@ static void stroke_timer(int id, float dt)
 
     float k;
 
+    if (button_state == BTN_CLICKED)
+    {
+        int elapsed = SDL_GetTicks() - button_time;
+        if (elapsed > DOUBLE_TAP_TIME)
+        {
+            /* it was a single click: start the simulation */
+            goto_state(&st_roll);
+            button_state = BTN_CLEAR;
+            return;
+        }
+    }
+
     if (SDL_GetModState() & KMOD_SHIFT ||
         (joystick && SDL_JoystickGetButton(joystick,
                                            config_get_d(CONFIG_JOYSTICK_BUTTON_B))))
@@ -928,6 +954,18 @@ static void stroke_timer(int id, float dt)
 
 static void stroke_point(int id, int x, int y, int dx, int dy)
 {
+    if (button_state == BTN_PRESS1 ||
+        button_state == BTN_PRESS2)
+    {
+        int duration = SDL_GetTicks() - button_time;
+        /* check how much we've moved */
+        if (abs(x - button_pressed_x) > TOUCH_SENSITIVITY ||
+            abs(y - button_pressed_y) > TOUCH_SENSITIVITY ||
+            duration > LONG_PRESS_TIME)
+            button_state = BTN_DRAGGING;
+    }
+    if (button_state != BTN_DRAGGING) return;
+
     game_set_rot(dx);
     game_set_mag(dy);
 }
@@ -942,7 +980,37 @@ static void stroke_stick(int id, int a, float v, int bump)
 
 static int stroke_click(int b, int d)
 {
-    return (d && b == SDL_BUTTON_LEFT) ? goto_state(&st_roll) : 1;
+    int ret = 1;
+    Uint32 current_time;
+
+    current_time = SDL_GetTicks();
+    if (d)
+    {
+        int y;
+        SDL_GetMouseState(&button_pressed_x, &y);
+        button_pressed_y = -y + config_get_d(CONFIG_HEIGHT);
+        button_time = current_time;
+        button_state = (button_state == BTN_CLICKED) ? BTN_PRESS2 : BTN_PRESS1;
+    }
+    else
+    {
+        /* is it a double click? */
+        int duration = current_time - button_time;
+        if (button_state == BTN_PRESS2)
+        {
+            button_state = BTN_CLEAR;
+            if (duration < DOUBLE_TAP_TIME)
+                ret = goto_pause(&st_over, 1);
+        }
+        else
+        {
+            if (button_state == BTN_PRESS1 && duration < LONG_PRESS_TIME)
+                button_state = BTN_CLICKED;
+            else
+                button_state = BTN_CLEAR;
+        }
+    }
+    return ret;
 }
 
 static int stroke_buttn(int b, int d)
