@@ -17,8 +17,12 @@
 #include "glext.h"
 #include "config.h"
 #include "syswm.h"
+#include "sync.h"
 
 /*---------------------------------------------------------------------------*/
+#ifdef __MAEMO__
+static SDL_GLES_Context *context;
+#endif
 
 int video_init(const char *title, const char *icon)
 {
@@ -34,6 +38,14 @@ int video_init(const char *title, const char *icon)
 
     set_SDL_icon(icon);
 
+#ifdef __MAEMO__
+    if (SDL_GLES_Init(SDL_GLES_VERSION_1_1) != 0)
+    {
+        fprintf(stderr, "SDL_GLES_Init failed\n");
+        return 0;
+    }
+#endif
+
     /* Initialize the video. */
 
     if (!video_mode(config_get_d(CONFIG_FULLSCREEN),
@@ -55,25 +67,6 @@ int video_init(const char *title, const char *icon)
 
 /*---------------------------------------------------------------------------*/
 
-int check_extension(const char *needle)
-{
-    const GLubyte *haystack, *c;
-
-    /* Search for the given string in the OpenGL extension strings. */
-
-    for (haystack = glGetString(GL_EXTENSIONS); *haystack; haystack++)
-    {
-        for (c = (const GLubyte *) needle; *c && *haystack; c++, haystack++)
-            if (*c != *haystack)
-                break;
-
-        if ((*c == 0) && (*haystack == ' ' || *haystack == '\0'))
-            return 1;
-    }
-
-    return 0;
-}
-
 int video_mode(int f, int w, int h)
 {
     int stereo  = config_get_d(CONFIG_STEREO)      ? 1 : 0;
@@ -90,22 +83,50 @@ int video_mode(int f, int w, int h)
 
     /* Require 16-bit double buffer with 16-bit depth buffer. */
 
+#ifndef __MAEMO__ /* SDL_GLES already takes care of these */
     SDL_GL_SetAttribute(SDL_GL_RED_SIZE,     5);
     SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,   5);
     SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,    5);
     SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,  16);
     SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
+#endif
 
     /* Try to set the currently specified mode. */
-
-    if (SDL_SetVideoMode(w, h, 0, SDL_OPENGL | (f ? SDL_FULLSCREEN : 0)))
+    if (SDL_SetVideoMode(w, h, 0,
+#ifdef __MAEMO__
+                         SDL_SWSURFACE |
+#else
+                         SDL_OPENGL |
+#endif
+                         (f ? SDL_FULLSCREEN : 0)))
     {
         config_set_d(CONFIG_FULLSCREEN, f);
         config_set_d(CONFIG_WIDTH,      w);
         config_set_d(CONFIG_HEIGHT,     h);
 
+#ifdef __MAEMO__
+        SDL_ShowCursor(SDL_DISABLE);
+        context = SDL_GLES_CreateContext();
+        if (context == 0)
+        {
+            fprintf(stderr, "SDL_GLES_CreateContext failed\n");
+            fprintf(stderr, "%s\n", SDL_GetError());
+            return 0;
+        }
+
+        if (SDL_GLES_MakeCurrent(context) != 0)
+        {
+            fprintf(stderr, "SDL_GLES_MakeCurrent failed\n");
+            fprintf(stderr, "%s\n", SDL_GetError());
+            return 0;
+        }
+#endif
+
+        if (!glext_init())
+            return 0;
+
         glViewport(0, 0, w, h);
-        glClearColor(0.0f, 0.0f, 0.1f, 0.0f);
+        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
 
         glEnable(GL_NORMALIZE);
         glEnable(GL_CULL_FACE);
@@ -114,25 +135,29 @@ int video_mode(int f, int w, int h)
         glEnable(GL_LIGHTING);
         glEnable(GL_BLEND);
 
+#if !ENABLE_OPENGLES
         glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL,
                       GL_SEPARATE_SPECULAR_COLOR);
-        glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
+#endif
+
+        glPixelStorei(GL_PACK_ALIGNMENT, 1);
+        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         glDepthFunc(GL_LEQUAL);
 
         /* If GL supports multisample, and SDL got a multisample buffer... */
 
-#ifdef GL_ARB_multisample
-        if (check_extension("ARB_multisample"))
+        if (glext_check("ARB_multisample"))
         {
             SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS, &buffers);
-            if (buffers)
-                glEnable(GL_MULTISAMPLE_ARB);
+            if (buffers) glEnable(GL_MULTISAMPLE);
         }
-#endif
 
-        glReadBuffer(GL_FRONT);
+        /* Attempt manual swap control if SDL's is broken. */
+
+        if (vsync && SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &vsync) == -1)
+            sync_init();
 
         return 1;
     }
@@ -236,16 +261,20 @@ void video_set_grab(int w)
         SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);
     }
 
+#ifndef __MAEMO__
     SDL_WM_GrabInput(SDL_GRAB_ON);
     SDL_ShowCursor(SDL_DISABLE);
+#endif
 
     grabbed = 1;
 }
 
 void video_clr_grab(void)
 {
+#ifndef __MAEMO__
     SDL_WM_GrabInput(SDL_GRAB_OFF);
     SDL_ShowCursor(SDL_ENABLE);
+#endif
     grabbed = 0;
 }
 
@@ -258,52 +287,52 @@ int  video_get_grab(void)
 
 void video_push_persp(float fov, float n, float f)
 {
-    GLdouble m[4][4];
+    GLfloat m[4][4];
 
-    GLdouble r = fov / 2 * V_PI / 180;
-    GLdouble s = sin(r);
-    GLdouble c = cos(r) / s;
+    GLfloat r = fov / 2 * V_PI / 180;
+    GLfloat s = sin(r);
+    GLfloat c = cos(r) / s;
 
-    GLdouble a = ((GLdouble) config_get_d(CONFIG_WIDTH) /
-                  (GLdouble) config_get_d(CONFIG_HEIGHT));
+    GLfloat a = ((GLfloat) config_get_d(CONFIG_WIDTH) /
+                 (GLfloat) config_get_d(CONFIG_HEIGHT));
 
     glMatrixMode(GL_PROJECTION);
     {
         glPushMatrix();
         glLoadIdentity();
 
-        m[0][0] =  c/a;
-        m[0][1] =  0.0;
-        m[0][2] =  0.0;
-        m[0][3] =  0.0;
-        m[1][0] =  0.0;
-        m[1][1] =    c;
-        m[1][2] =  0.0;
-        m[1][3] =  0.0;
-        m[2][0] =  0.0;
-        m[2][1] =  0.0;
+        m[0][0] = c / a;
+        m[0][1] =  0.0f;
+        m[0][2] =  0.0f;
+        m[0][3] =  0.0f;
+        m[1][0] =  0.0f;
+        m[1][1] =     c;
+        m[1][2] =  0.0f;
+        m[1][3] =  0.0f;
+        m[2][0] =  0.0f;
+        m[2][1] =  0.0f;
         m[2][2] = -(f + n) / (f - n);
-        m[2][3] = -1.0;
-        m[3][0] =  0.0;
-        m[3][1] =  0.0;
-        m[3][2] = -2.0 * n * f / (f - n);
-        m[3][3] =  0.0;
+        m[2][3] = -1.0f;
+        m[3][0] =  0.0f;
+        m[3][1] =  0.0f;
+        m[3][2] = -2.0f * n * f / (f - n);
+        m[3][3] =  0.0f;
 
-        glMultMatrixd(&m[0][0]);
+        glMultMatrixf(&m[0][0]);
     }
     glMatrixMode(GL_MODELVIEW);
 }
 
 void video_push_ortho(void)
 {
-    GLdouble w = (GLdouble) config_get_d(CONFIG_WIDTH);
-    GLdouble h = (GLdouble) config_get_d(CONFIG_HEIGHT);
+    GLfloat w = (GLfloat) config_get_d(CONFIG_WIDTH);
+    GLfloat h = (GLfloat) config_get_d(CONFIG_HEIGHT);
 
     glMatrixMode(GL_PROJECTION);
     {
         glPushMatrix();
         glLoadIdentity();
-        glOrtho(0.0, w, 0.0, h, -1.0, +1.0);
+        glOrtho_(0.0, w, 0.0, h, -1.0, +1.0);
     }
     glMatrixMode(GL_MODELVIEW);
 }
@@ -320,12 +349,10 @@ void video_pop_matrix(void)
 void video_clear(void)
 {
     if (config_get_d(CONFIG_REFLECTION))
-        glClear(GL_COLOR_BUFFER_BIT |
-                GL_DEPTH_BUFFER_BIT |
+        glClear(GL_DEPTH_BUFFER_BIT |
                 GL_STENCIL_BUFFER_BIT);
     else
-        glClear(GL_COLOR_BUFFER_BIT |
-                GL_DEPTH_BUFFER_BIT);
+        glClear(GL_DEPTH_BUFFER_BIT);
 }
 
 /*---------------------------------------------------------------------------*/
